!**********************************************************************
!   poisson.f90 - Resolution de l'equation de Poisson en utilisant Jacobi
!   sur le domaine [0,1]x[0,1]x[0,1]par une methode aux differences
!   finies, avec comme solveur Jacobi.
!   Delta u = f(x,y)= 2yz(y-1)(z-1)+2xz(x-1)(z-1)+2xy(x-1)(y-1)
!   u sur les bords vaut 0
!   La solution exacte est u= xyz(x-1)(y-1)(z-1)
!
!   La valeur de u est donnee par la formule:
!   u(i,j,k)(n+1) = 1/6.*(u(i+1,j,k)(n)+u(i-1,j,k)(n) +
!                         u(i,j+1,k)(n)+u(i,j-1,k)(n) +
!                         u(i,j,k+1)(n)+u(i,j,k-1)(n) - h*h*f(i,j))
!
!   Dans cette version, on se donne le nombre de points interieurs
!   total suivant x (ntx), y (nty) et z (ntz).
!   ntx = nty = ntz
!
!   hx represente le pas suivant x, hy le pas suivant y, et hz suivant z
!   comme ntx = nty = ntz , hx=hy=hz=h=1/(ntx+1)
!
!   Pour chaque processus :
!   1) decomposer le domaine
!   2) connaitre ses 6 voisins
!   3) echanger les points aux interfaces
!   4) calculer
!
!   Auteur          : Isabelle DUPAYS (CNRS/IDRIS - France)
!                     <Isabelle.Dupays@idris.fr>
!   Cree le         : Fri Nov 15 10:32:45 1996
!   Modifie par     : Philippe WAUTELET (CNRS/IDRIS - France)
!                     <Philippe.Wautelet@idris.fr>
!****************************************************************************
PROGRAM poisson
  USE mod_poisson
  use mod_calcul
  use mod_comm
  use mod_dom
  use mod_erreur
  use mod_init
  use mod_vois

  IMPLICIT  NONE

  !---Declaration des variables

  !Convergence
  REAL(kind=8)                              :: diffnorm, dwork
  !Mesure du temps
  REAL(kind=8)                              :: t1, t2
  !Compteur
  INTEGER                                   :: j
  !Test de convergence
  LOGICAL                                   :: convergence
  !Argument muet de la fonction F90 EPSILON
  REAL(kind=8)                              :: x
  !Types derives utilises pour les communications sur les faces du domaine
  INTEGER                                   :: plan_xy, plan_yz, plan_xz
  !Variables intermediaires
  INTEGER                                   :: ligne_y, size_of_double
  INTEGER(KIND=MPI_ADDRESS_KIND)            :: step_size

  LOGICAL, DIMENSION(ndims)                 :: periods
  LOGICAL, PARAMETER                        :: reorganisation = .false.

  !****************************************************************************

  !Initialisation de MPI
  CALL MPI_INIT(code)

  !Savoir quel processus je suis
  CALL MPI_COMM_RANK( MPI_COMM_WORLD, rang, code)

  !Connaitre le nombre total de processus
  CALL MPI_COMM_SIZE( MPI_COMM_WORLD, nb_procs, code)

  !Pour le processus 0, connaitre le nombre de points interieurs total en x
  IF (rang == 0) THEN
     ntx = taille_probleme
     print *,'Taille probleme: ',ntx,'x',ntx,'x',ntx
  END IF

  !Envoi de la valeur de ntx a tous les autres processus
  CALL MPI_BCAST(ntx, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, code)

  !Determiner le nombre de points interieurs total en y nty = ntx
  nty = ntx
  ntz = ntx

  !Connaitre le nombre de processus selon x et le nombre de processus
  !selon y en fonction du nombre total de processus
  dims(:)= 0
  CALL MPI_DIMS_CREATE( nb_procs, ndims, dims, code )
  IF (rang==0) print *,'Topo:',dims

  !Creation de la grille de processus 3D sans periodicite
  periods(:) = .false.

  CALL MPI_CART_CREATE( MPI_COMM_WORLD, ndims, dims, periods, &
                        reorganisation, comm3d, code )

  !Determiner les indices de chaque sous domaine
  CALL domaine()

  !Allocation dynamique des tableaux u,u_nouveau,u_exact,f
  ALLOCATE (u(sx-1:ex+1,sy-1:ey+1,sz-1:ez+1),u_nouveau(sx-1:ex+1,sy-1:ey+1,sz-1:ez+1))
  ALLOCATE (f(sx-1:ex+1,sy-1:ey+1,sz-1:ez+1),u_exact(sx-1:ex+1,sy-1:ey+1,sz-1:ez+1))

  !Initialisation du second membre et de la solution u sur les bords
  CALL initialisation()

  !Recherche de ses 6 voisins pour chaque processus
  CALL voisinage()

  !PRINT *, "Processus ", rang, ":",voisin

  !Creation des types derives
  CALL MPI_TYPE_VECTOR(ey-sy+1,ex-sx+1,ex-sx+3,MPI_DOUBLE_PRECISION,plan_xy,code)
  CALL MPI_TYPE_COMMIT(plan_xy,code)

  CALL MPI_TYPE_VECTOR(ez-sz+1,ex-sx+1,(ex-sx+3)*(ey-sy+3),MPI_DOUBLE_PRECISION,plan_xz,code)
  CALL MPI_TYPE_COMMIT(plan_xz,code)

  CALL MPI_TYPE_SIZE (MPI_DOUBLE_PRECISION,size_of_double,code)
  step_size = (ex-sx+3)*(ey-sy+3)*size_of_double
  CALL MPI_TYPE_VECTOR(ey-sy+1,1,ex-sx+3,MPI_DOUBLE_PRECISION,ligne_y,code)
  CALL MPI_TYPE_CREATE_HVECTOR(ez-sz+1,1,step_size,ligne_y,plan_yz,code)
  CALL MPI_TYPE_COMMIT(plan_yz,code)

  !Schema iteratif en temps
  it = 0
  convergence = .false.

  !Mesure du temps en seconde dans la boucle en temps
  t1 = MPI_WTIME()

  DO WHILE ((.NOT. convergence) .AND. (it < it_max))

     it = it +1

     u(:,:,:) = u_nouveau(:,:,:)

     !Echange des points aux interfaces pour u a l'iteration n
     CALL communication(plan_xy, plan_yz, plan_xz)

     !Calcul de u a l'iteration n+1
     CALL calcul()

     !Calcul de l'erreur sur le sous-domaine
     dwork = erreur()

     !Calcul de l'erreur sur tous les sous-domaines
     CALL MPI_AllREDUCE( dwork, diffnorm, 1, MPI_DOUBLE_PRECISION, &
                         MPI_MAX, comm3d, code )

     !Arret du programme si on a atteint 100 fois la precision machine
     !obtenu par la fonction F90 EPSILON
     convergence = (diffnorm < 100*EPSILON(X))

     !Affichage pour le processus 0 de la difference
     IF ((rang == 0) .AND.  (MOD(it,1000) == 0) )THEN
        PRINT *, it, ' La difference est ', diffnorm
     END IF

  END DO

  !Mesure du temps a la sortie de la boucle
  t2 = MPI_WTIME()

  !Affichage du temps de convergence par le processus 0
  IF (rang ==  0) THEN
     PRINT *, 'Convergence apres ', it, ' Iterations en  ', t2 - t1, ' secs '
  END IF

  !Comparaison de la solution calculee et de la solution exacte
  !sur le processus 0
  IF (rang ==  0) THEN
     PRINT *,'Solution exacte u_exact ', 'Solution calculee u'
     DO j=sy,ey,(ey-sy)/10
        PRINT 10, u_exact(1,j,1), u (1,j,1)
10      FORMAT  ('u_exact=  ',E12.5,' u =  ',E12.5)
     END DO
  END IF

  !Desallocation des tableaux u,u_nouveau,u_exact et f
  DEALLOCATE(u, u_nouveau, u_exact, f)

  !Liberation des types derives et du communicateur comm3d
  CALL MPI_TYPE_FREE( plan_xy, code )
  CALL MPI_TYPE_FREE( plan_xz, code )
  CALL MPI_TYPE_FREE( plan_yz, code )
  CALL MPI_COMM_FREE( comm3d, code )

  !Desactivation de MPI
  CALL MPI_FINALIZE(code)

END PROGRAM poisson
